import json
from typing import List, Optional


def _run(pydantic_version):
    if pydantic_version == "v1":
        from pydantic.v1 import Field

        # based on pydantic v1
        from oold.model.v1 import (
            LinkedBaseModel,
            ResolveParam,
            Resolver,
            ResolveResult,
            SetResolverParam,
            set_resolver,
        )

        class Entity(LinkedBaseModel):
            class Config:
                schema_extra = {
                    "@context": {
                        # aliases
                        "id": "@id",
                        "type": "@type",
                        # prefixes
                        "schema": "https://schema.org/",
                        "ex": "https://example.com/",
                        # literal property
                        "name": "schema:name",
                    },
                    "iri": "Entity.json",  # the IRI of the schema
                }

            type: Optional[str] = "ex:Entity.json"
            name: str

            def get_iri(self):
                return "ex:" + self.name

        class Person(Entity):
            class Config:
                schema_extra = {
                    "@context": [
                        "Entity.json",  # import the context of the parent class
                        {
                            # object property definition
                            "knows": {
                                "@id": "schema:knows",
                                "@type": "@id",
                                "@container": "@set",
                            }
                        },
                    ],
                    "iri": "Person.json",
                }

            type: Optional[str] = "ex:Person.json"
            knows: Optional[List["Person"]] = Field(
                None,
                # object property pointing to another Person
                range="Person.json",
            )

    if pydantic_version == "v2":
        from pydantic import ConfigDict, Field

        # based on pydantic v2
        from oold.model import LinkedBaseModel  # noqa
        from oold.model import ResolveParam  # noqa
        from oold.model import Resolver  # noqa
        from oold.model import ResolveResult  # noqa
        from oold.model import SetResolverParam  # noqa
        from oold.model import set_resolver  # noqa

        class Entity(LinkedBaseModel):
            model_config = ConfigDict(
                json_schema_extra={
                    "@context": {
                        # aliases
                        "id": "@id",
                        "type": "@type",
                        # prefixes
                        "schema": "https://schema.org/",
                        "ex": "https://example.com/",
                        # literal property
                        "name": "schema:name",
                    },
                    "iri": "Entity.json",  # the IRI of the schema
                }
            )
            type: Optional[str] = "ex:Entity.json"
            name: str

            def get_iri(self):
                return "ex:" + self.name

        class Person(Entity):  # noqa
            model_config = ConfigDict(
                json_schema_extra={
                    "@context": [
                        "Entity.json",  # import the context of the parent class
                        {
                            # object property definition
                            "knows": {
                                "@id": "schema:knows",
                                "@type": "@id",
                                "@container": "@set",
                            }
                        },
                    ],
                    "iri": "Person.json",
                }
            )
            type: Optional[str] = "ex:Person.json"
            knows: Optional[List["Person"]] = Field(
                None,
                # object property pointing to another Person
                json_schema_extra={"range": "Person.json"},
            )

    p1 = Person(name="Alice")
    p2 = Person(name="Bob", knows=[p1])
    print(p2.to_jsonld())

    # load the rdf into a rdflib graph
    from rdflib import Graph

    g = Graph()
    g.parse(data=p1.to_jsonld(), format="json-ld")
    g.parse(data=p2.to_jsonld(), format="json-ld")
    print(g.serialize(format="turtle"))

    # query the name of persons that Bob knows
    qres = g.query(
        """
        SELECT ?name
        WHERE {
            ?s <https://schema.org/knows> ?o .
            ?o <https://schema.org/name> ?name .
        }
        """
    )
    for row in qres:
        print("Bob knows " + row.name)
        assert str(row.name) == "Alice"

    # create a resolver to resolve IRIs to objects
    class SparqlResolver(Resolver):
        # model_config = ConfigDict(
        #    arbitrary_types_allowed=True
        # )
        class Config:
            arbitrary_types_allowed = True

        graph: Graph

        def resolve_iri(self, iri):
            # sparql query to get a node by IRI with all its properties
            # using CONSTRUCT to get the full node
            # format the result as json-ld
            iri_filter = f"FILTER (?s = {iri})"
            # check if the iri is a full IRI or a prefix
            if iri.startswith("http"):
                iri_filter = f"FILTER (?s = <{iri}>)"
            qres = self.graph.query(
                """
                PREFIX ex: <https://example.com/>
                CONSTRUCT {
                    ?s ?p ?o .
                }
                WHERE {
                    ?s ?p ?o .
                    {{{iri_filter}}}
                }
                """.replace(
                    "{{{iri_filter}}}", iri_filter
                )
            )
            jsonld_dict = json.loads(qres.serialize(format="json-ld"))[0]
            res = LinkedBaseModel.from_jsonld(jsonld_dict)
            return res

        def resolve(self, request: ResolveParam):
            # print("RESOLVE", request)
            nodes = {}
            for iri in request.iris:
                nodes[iri] = self.resolve_iri(iri)
            return ResolveResult(nodes=nodes)

    r = SparqlResolver(graph=g)
    set_resolver(SetResolverParam(iri="ex", resolver=r))

    # load bob from the graph
    bob = r.resolve_iri("ex:Bob")
    print(bob)
    # accessing 'knows' will trigger a sparql query
    # to get the full node of Alice from the graph
    print(bob.knows[0].name)
    assert bob.knows[0].name == "Alice"


def test_rdf_export_and_sparql_query():
    _run("v1")
    _run("v2")


if __name__ == "__main__":
    test_rdf_export_and_sparql_query()



"""Tests for `oold` package."""

import importlib
import json
from typing import Any

import datamodel_code_generator

import oold.model.model as model
from oold.generator import Generator


def _run(pydantic_version="v1"):
    if pydantic_version == "v1":
        from oold.model.v1 import (
            ResolveParam,
            Resolver,
            ResolveResult,
            SetResolverParam,
            set_resolver,
        )

        output_model_type = datamodel_code_generator.DataModelType.PydanticBaseModel
    else:
        from oold.model import (
            ResolveParam,
            Resolver,
            ResolveResult,
            SetResolverParam,
            set_resolver,
        )

        output_model_type = datamodel_code_generator.DataModelType.PydanticV2BaseModel

    """Tests for `oold` package."""

    schemas = [
        {
            "id": "./bar2/Bar2",
            "title": "Bar2",
            "type": "object",
            "properties": {
                "id": {"type": "string"},
                "type": {
                    "type": "array",
                    "items": {"type": "string"},
                    "default": ["Bar2"],
                },
                "prop1": {"type": "string"},
            },
        },
        {
            "id": "Bar",
            "title": "Bar",
            "type": "object",
            "allOf": [{"$ref": "./bar2/Bar2.json"}],
            "properties": {
                "type": {
                    "type": "array",
                    "items": {"type": "string"},
                    "default": ["Bar"],
                },
                "prop2": {"type": "string"},
            },
        },
        {
            "id": "Foo",
            "title": "Foo",
            "type": "object",
            "required": ["id", "b"],
            "properties": {
                "id": {"type": "string"},
                "type": {
                    "type": "array",
                    "items": {"type": "string"},
                    "default": ["Foo"],
                },
                "literal": {"type": "string"},
                "b": {"type": "string", "range": "Bar.json"},
                "b_default": {"type": "string", "range": "Bar.json", "default": "ex:b"},
                "b2": {
                    "type": "array",
                    "items": {"type": "string", "range": "Bar.json"},
                },
            },
        },
    ]
    graph = [
        {
            "id": "ex:f",
            "type": ["Foo"],
            "literal": "test1",
            "b": "ex:b",
            # will be automatically added by the class constructor
            # "b_default": "ex:b",
            "b2": ["ex:b1", "ex:b2"],
        },
        {"id": "ex:b", "type": ["Bar"], "prop1": "test2"},
        {"id": "ex:b1", "type": ["Bar"], "prop1": "test3"},
        {"id": "ex:b2", "type": ["Bar"], "prop1": "test4"},
    ]

    g = Generator()
    g.generate(schemas, main_schema="Foo.json", output_model_type=output_model_type)
    importlib.reload(model)

    class MyResolver(Resolver):
        graph: (Any)

        def resolve_iri(self, iri):
            for node in self.graph:
                if node["id"] == iri:
                    cls = node["type"][0]
                    entity = eval(f"model.{cls}(**node)")
                    return entity

        def resolve(self, request: ResolveParam):
            # print("RESOLVE", request)
            nodes = {}
            for iri in request.iris:
                nodes[iri] = self.resolve_iri(iri)
            return ResolveResult(nodes=nodes)

    r = MyResolver(graph=graph)
    set_resolver(SetResolverParam(iri="ex", resolver=r))

    f = model.Foo(id="ex:f", literal="test1", b="ex:b", b2=["ex:b1", "ex:b2"])
    print(f.b)

    print(f.b.id)
    assert f.b.id == "ex:b"
    for b in f.b2:
        print(b)
    assert f.b2[0].id == "ex:b1" and f.b2[0].prop1 == "test3"
    assert f.b2[1].id == "ex:b2" and f.b2[1].prop1 == "test4"
    assert f.b_default.id == "ex:b"

    f = model.Foo(
        id="ex:f",
        literal="test1",
        b=model.Bar(id="ex:b", prop1="test2"),
        b2=[model.Bar(id="ex:b1", prop1="test3"), model.Bar(id="ex:b2", prop1="test4")],
    )
    assert f.b.id == "ex:b"
    for b in f.b2:
        print(b)
    assert f.b2[0].id == "ex:b1" and f.b2[0].prop1 == "test3"
    assert f.b2[1].id == "ex:b2" and f.b2[1].prop1 == "test4"

    def export_json(obj):
        if pydantic_version == "v1":
            return obj.json(exclude_none=True)
        return obj.model_dump_json(exclude_none=True)

    print(export_json(f))
    assert json.loads(export_json(f)) == {**graph[0], **{"b_default": "ex:b"}}
    assert json.loads(export_json(f.b)) == graph[1]
    assert json.loads(export_json(f.b2[0])) == graph[2]
    assert json.loads(export_json(f.b2[1])) == graph[3]

    # Test nonexisting IRIs => properties should be initialized to None
    # but IRI is persisted when exporting to JSON
    f = model.Foo(
        id="ex:f",
        literal="test1",
        b="ex:b",
        b_default="ex:doesNotExist",
        b2=["ex:b1", "ex:doesNotExist"],
    )

    print(f)
    print(export_json(f))
    assert f.b_default is None
    assert f.b2[1] is None
    f_json = json.loads(export_json(f))
    assert f_json["b_default"] == "ex:doesNotExist"
    assert f_json["b2"][1] == "ex:doesNotExist"

    # test if skipping of a required property throws an exception
    # assert that ValueError is raised, fail if non is raised
    try:
        f = model.Foo(
            id="ex:f",
            literal="test1",
            # b="ex:b",
            b_default="ex:doesNotExist",
            b2=["ex:b1", "ex:doesNotExist"],
        )
    except Exception as e:
        assert isinstance(e, ValueError)
    else:
        assert False, "ValueError not raised"


def test_core():
    _run(pydantic_version="v1")
    _run(pydantic_version="v2")


if __name__ == "__main__":
    test_core()



import oold.model.model as model

from .static import _run


def test_oneof_subschema():
    # json schema with property that contains a oneOf with two subschemas

    schemas = [
        {
            "id": "example",
            "title": "Example",
            "type": "object",
            "properties": {
                "type": {"type": "string", "default": ["example"]},
                "prop1": {"type": "string", "custom_key": "custom_value"},
                "prop2": {
                    "custom_key": "custom_value",
                    "properties": {
                        "subprop0": {"type": "string", "custom_key": "custom_value_0"},
                    },
                    "oneOf": [
                        {
                            "title": "Subschema1",
                            "type": "object",
                            "properties": {
                                "subprop1": {
                                    "type": "string",
                                    "custom_key": "custom_value_1",
                                },
                            },
                        },
                        {
                            "title": "Subschema2",
                            "type": "object",
                            "properties": {
                                "subprop2": {
                                    "type": "string",
                                    "custom_key": "custom_value_2",
                                },
                            },
                        },
                    ],
                },
            },
        },
    ]

    def test(pydantic_version):
        # Test the generated model, see
        # https://github.com/koxudaxi/datamodel-code-generator/issues/2403

        if pydantic_version == "v1":
            assert (
                model.Subschema1.__fields__["subprop1"].field_info.extra["custom_key"]
                == "custom_value_1"
            )
        else:
            model.Subschema1.model_fields["subprop1"].json_schema_extra[
                "custom_key"
            ] == "custom_value_1"

    _run(
        schemas,
        main_schema="example.json",
        test=test,
        # pydantic_versions=["v1"],
    )


if __name__ == "__main__":
    test_oneof_subschema()



import importlib
from typing import Callable, List

import datamodel_code_generator

import oold.model.model as model
from oold.generator import Generator


def _run(
    schemas: List, main_schema: str, test: Callable, pydantic_versions=["v1", "v2"]
):
    for pydantic_version in pydantic_versions:
        if pydantic_version == "v1":
            # from oold.model.v1 import (
            #     ResolveParam,
            #     Resolver,
            #     ResolveResult,
            #     SetResolverParam,
            #     set_resolver,
            # )

            output_model_type = datamodel_code_generator.DataModelType.PydanticBaseModel
        else:
            # from oold.model import (
            #     ResolveParam,
            #     Resolver,
            #     ResolveResult,
            #     SetResolverParam,
            #     set_resolver,
            # )

            output_model_type = (
                datamodel_code_generator.DataModelType.PydanticV2BaseModel
            )

        g = Generator()
        g.generate(
            schemas, main_schema=main_schema, output_model_type=output_model_type
        )
        importlib.reload(model)

        test(pydantic_version)



import json
import os
import re
from pathlib import Path
from pprint import pprint
from tempfile import TemporaryDirectory
from typing import List, Optional

from datamodel_code_generator import DataModelType, InputFileType, generate
from datamodel_code_generator.parser.jsonschema import (
    JsonSchemaObject,
    JsonSchemaParser,
)


def generate1(json_schemas):
    code = ""
    first = True
    for schema in json_schemas:
        parser = JsonSchemaParser(
            json.dumps(schema),
            # custom_template_dir=Path(model_dir_path),
            field_include_all_keys=True,
            base_class="osw.model.static.OswBaseModel",
            # use_default = True,
            enum_field_as_literal="all",
            use_title_as_name=True,
            use_schema_description=True,
            use_field_description=True,
            encoding="utf-8",
            use_double_quotes=True,
            collapse_root_models=True,
            reuse_model=True,
        )
        content = parser.parse()

        if first:
            header = (
                "from uuid import uuid4\n"
                "from typing import Type, TypeVar\n"
                "from osw.model.static import OswBaseModel, Ontology\n"
                "\n"
            )

            content = re.sub(
                r"(class\s*\S*\s*\(\s*OswBaseModel\s*\)\s*:.*\n)",
                header + r"\n\n\n\1",
                content,
                1,
            )  # add header before first class declaration

            content = re.sub(
                r"(UUID = Field\(...)",
                r"UUID = Field(default_factory=uuid4",
                content,
            )  # enable default value for uuid

        else:
            org_content = code

            pattern = re.compile(
                r"class\s*([\S]*)\s*\(\s*\S*\s*\)\s*:.*\n"
            )  # match class definition [\s\S]*(?:[^\S\n]*\n){2,}
            for cls in re.findall(pattern, org_content):
                print(cls)
                content = re.sub(
                    r"(class\s*"
                    + cls
                    + r"\s*\(\s*\S*\s*\)\s*:.*\n[\s\S]*?(?:[^\S\n]*\n){3,})",
                    "",
                    content,
                    count=1,
                )  # replace duplicated classes

            content = re.sub(
                r"(from __future__ import annotations)", "", content, 1
            )  # remove import statement

        code += content + "\r\n"
        # pprint(parser.raw_obj)
        # print(result)
        first = False

    with open("model.py", "w") as f:
        f.write(code)


def generate2(json_schemas):
    with TemporaryDirectory() as temporary_directory_name:
        temporary_directory = Path(temporary_directory_name)
        temporary_directory = Path("./model")
        output = Path(temporary_directory / "model.py")
        for schema in json_schemas:
            name = schema["id"]
            with open(
                Path(temporary_directory / (name + ".json")), "w", encoding="utf-8"
            ) as f:
                schema_str = json.dumps(schema, ensure_ascii=False, indent=4).replace(
                    "dollarref", "$ref"
                )
                # print(schema_str)
                f.write(schema_str)
        generate(
            input_=Path(temporary_directory / "Foo.json"),
            # json_schema,
            input_file_type=InputFileType.JsonSchema,
            input_filename="Foo.json",
            output=output,
            # set up the output model types
            output_model_type=DataModelType.PydanticV2BaseModel,
            # custom_template_dir=Path(model_dir_path),
            field_include_all_keys=True,
            base_class="static.LinkedBaseModel",
            # use_default = True,
            enum_field_as_literal="all",
            use_title_as_name=True,
            use_schema_description=True,
            use_field_description=True,
            encoding="utf-8",
            use_double_quotes=True,
            collapse_root_models=True,
            reuse_model=True,
        )


def get_schemas():
    json_schemas = [
        {
            "id": "Bar",
            "title": "Bar",
            "type": "object",
            "properties": {
                "type": {
                    "type": "array",
                    "items": {"type": "string"},
                    "default": ["Bar"],
                },
                "prop1": {"type": "string"},
            },
        },
        {
            "id": "Foo",
            "title": "Foo",
            "type": "object",
            "properties": {
                "type": {
                    "type": "array",
                    "items": {"type": "string"},
                    "default": ["Foo"],
                },
                "literal": {"type": "string"},
                "b": {"type": "string", "range": "Bar"},
                "b2": {"$ref": "Bar.json", "range": "Bar"},
            },
        },
    ]
    return json_schemas


def preprocess(json_schemas):
    aggr_schema = {"$defs": {}}
    for schema in json_schemas:
        aggr_schema["$defs"][schema["id"]] = schema
    # pprint(aggr_schema)
    return aggr_schema


js = get_schemas()
pprint(js)
# js = preprocess(js)
generate2(js)



import json
from abc import abstractmethod
from typing import Any, Dict, List, Literal, Optional, Union

import pydantic
from pydantic import BaseModel
from typing_extensions import Self  # to support python < 3.11

from oold.model.static import GenericLinkedBaseModel, export_jsonld, import_jsonld


class SetResolverParam(BaseModel):
    iri: str
    resolver: "Resolver"


class GetResolverParam(BaseModel):
    iri: str


class GetResolverResult(BaseModel):
    resolver: "Resolver"


class ResolveParam(BaseModel):
    iris: List[str]


class ResolveResult(BaseModel):
    nodes: Dict[str, Union[None, "LinkedBaseModel"]]


class Resolver(BaseModel):
    @abstractmethod
    def resolve(self, request: ResolveParam) -> ResolveResult:
        pass


global _resolvers
_resolvers = {}


def set_resolver(param: SetResolverParam) -> None:
    _resolvers[param.iri] = param.resolver


def get_resolver(param: GetResolverParam) -> GetResolverResult:
    # ToDo: Handle prefixes (ex:) as well as full IRIs (http://example.com/)
    iri = param.iri.split(":")[0]
    if iri not in _resolvers:
        raise ValueError(f"No resolvers found for {iri}")
    return GetResolverResult(resolver=_resolvers[iri])


# pydantic v2
_types: Dict[str, pydantic.main._model_construction.ModelMetaclass] = {}


# pydantic v2
class LinkedBaseModelMetaClass(pydantic.main._model_construction.ModelMetaclass):
    def __new__(mcs, name, bases, namespace):
        cls = super().__new__(mcs, name, bases, namespace)
        schema = {}

        # pydantic v2
        if "model_config" in namespace:
            if "json_schema_extra" in namespace["model_config"]:
                schema = namespace["model_config"]["json_schema_extra"]

        if "iri" in schema:
            iri = schema["iri"]
            _types[iri] = cls
        return cls


class LinkedBaseModel(
    BaseModel, GenericLinkedBaseModel, metaclass=LinkedBaseModelMetaClass
):
    """LinkedBaseModel for pydantic v2"""

    __iris__: Optional[Dict[str, Union[str, List[str]]]] = {}

    def get_iri(self) -> str:
        """Return the unique IRI of the object.
        Overwrite this method in the subclass."""
        return self.id

    @classmethod
    def model_validate(
        cls,
        obj: Any,
        *,
        strict: Union[bool, None] = None,
        from_attributes: Union[bool, None] = None,
        context: Union[Any, None] = None,
    ) -> Self:
        """Validate a pydantic model instance.

        Args:
            obj: The object to validate.
            strict: Whether to enforce types strictly.
            from_attributes: Whether to extract data from object attributes.
            context: Additional context to pass to the validator.

        Raises:
            ValidationError: If the object could not be validated.

        Returns:
            The validated model instance.
        """
        if isinstance(obj, str):
            return cls._resolve([obj])[obj]
        if isinstance(obj, list):
            node_dict = cls._resolve(obj)
            node_list = []
            for iri in obj:
                node = node_dict[iri]
                if node:
                    node_list.append(node)
            return node_list
        elif isinstance(obj, dict):
            super().model_validate(
                obj, strict=strict, from_attributes=from_attributes, context=context
            )

    def __init__(self, *a, **kw):
        if "__iris__" not in kw:
            kw["__iris__"] = {}

        for name in list(kw):  # force copy of keys for inline-delete
            if name == "__iris__":
                continue
            if name not in self.model_fields:
                continue
            # rewrite <attr> to <attr>_iri
            # pprint(self.__fields__)
            extra = None
            # pydantic v1
            # if name in self.__fields__:
            #     if hasattr(self.__fields__[name].default, "json_schema_extra"):
            #         extra = self.__fields__[name].default.json_schema_extra
            #     elif hasattr(self.__fields__[name].field_info, "extra"):
            #         extra = self.__fields__[name].field_info.extra
            # pydantic v2
            extra = self.model_fields[name].json_schema_extra

            if extra and "range" in extra:
                arg_is_list = isinstance(kw[name], list)

                # annotation_is_list = False
                # args = self.model_fields[name].annotation.__args__
                # if hasattr(args[0], "_name"):
                #    is_list = args[0]._name == "List"
                if arg_is_list:
                    kw["__iris__"][name] = []
                    for e in kw[name][:]:  # interate over copy of list
                        if isinstance(e, BaseModel):  # contructed with object ref
                            kw["__iris__"][name].append(e.get_iri())
                        elif isinstance(e, str):  # constructed from json
                            kw["__iris__"][name].append(e)
                            kw[name].remove(e)  # remove to construct valid instance
                    if len(kw[name]) == 0:
                        # pydantic v1
                        # kw[name] = None # else pydantic v1 will set a FieldInfo object
                        # pydantic v2
                        kw[name] = None  # else default value may be set
                else:
                    if isinstance(kw[name], BaseModel):  # contructed with object ref
                        # print(kw[name].id)
                        kw["__iris__"][name] = kw[name].get_iri()
                    elif isinstance(kw[name], str):  # constructed from json
                        kw["__iris__"][name] = kw[name]
                        # pydantic v1
                        # kw[name] = None # else pydantic v1 will set a FieldInfo object
                        # pydantic v2
                        kw[name] = None  # else default value may be set

        BaseModel.__init__(self, *a, **kw)
        # handle default values
        for name in list(self.__dict__.keys()):
            if self.__dict__[name] is None:
                continue
            extra = None
            # pydantic v1
            # if name in self.__fields__:
            #     if hasattr(self.__fields__[name].default, "json_schema_extra"):
            #         extra = self.__fields__[name].default.json_schema_extra
            #     elif hasattr(self.__fields__[name].field_info, "extra"):
            #         extra = self.__fields__[name].field_info.extra
            # pydantic v2
            extra = self.model_fields[name].json_schema_extra

            if extra and "range" in extra:
                arg_is_list = isinstance(self.__dict__, list)

                if arg_is_list:
                    kw["__iris__"][name] = []
                    for e in self.__dict__[name]:
                        if isinstance(e, BaseModel):  # contructed with object ref
                            kw["__iris__"][name].append(e.get_iri())
                else:
                    if isinstance(
                        self.__dict__[name], BaseModel
                    ):  # contructed with object ref
                        kw["__iris__"][name] = self.__dict__[name].get_iri()

        self.__iris__ = kw["__iris__"]

        # iterate over all fields
        # if x-oold-required-iri occurs in extra and the field is not set in __iri__
        # throw an error
        for name in self.model_fields:
            extra = None
            # pydantic v1
            # if name in self.__fields__:
            #     if hasattr(self.__fields__[name].default, "json_schema_extra"):
            #         extra = self.__fields__[name].default.json_schema_extra
            #     elif hasattr(self.__fields__[name].field_info, "extra"):
            #         extra = self.__fields__[name].field_info.extra
            # pydantic v2
            extra = self.model_fields[name].json_schema_extra

            if extra and "x-oold-required-iri" in extra:
                if name not in self.__iris__:
                    raise ValueError(f"{name} is required but not set")

    def __getattribute__(self, name):
        # print("__getattribute__ ", name)
        # async? https://stackoverflow.com/questions/33128325/
        # how-to-set-class-attribute-with-await-in-init

        if name in ["__dict__", "__pydantic_private__", "__iris__"]:
            return BaseModel.__getattribute__(self, name)  # prevent loop

        else:
            if hasattr(self, "__iris__"):
                if name in self.__iris__ and len(self.__iris__[name]) > 0:
                    if self.__dict__[name] is None or (
                        isinstance(self.__dict__[name], list)
                        and len(self.__dict__[name]) == 0
                    ):
                        iris = self.__iris__[name]
                        is_list = isinstance(iris, list)
                        if not is_list:
                            iris = [iris]

                        node_dict = self._resolve(iris)
                        if is_list:
                            node_list = []
                            for iri in iris:
                                node = node_dict[iri]
                                node_list.append(node)
                            self.__setattr__(name, node_list)
                        else:
                            node = node_dict[iris[0]]
                            if node:
                                self.__setattr__(name, node)

        return BaseModel.__getattribute__(self, name)

    def model_dump(self, **kwargs):  # extent BaseClass export function
        # print("dict")
        remove_none = kwargs.get("remove_none", False)
        kwargs["exclude_none"] = False
        d = super().model_dump(**kwargs)
        # pprint(d)
        self._object_to_iri(d)
        if remove_none:
            d = self.remove_none(d)
        # pprint(d)
        return d

    @staticmethod
    def _resolve(iris):
        resolver = get_resolver(GetResolverParam(iri=iris[0])).resolver
        node_dict = resolver.resolve(ResolveParam(iris=iris)).nodes
        return node_dict

    # pydantic v2
    def model_dump_json(
        self,
        *,
        indent: Union[int, None] = None,
        include: Union[pydantic.main.IncEx, None] = None,
        exclude: Union[pydantic.main.IncEx, None] = None,
        context: Union[Any, None] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: Union[bool, Literal["none", "warn", "error"]] = True,
        serialize_as_any: bool = False,
        **dumps_kwargs: Any,
    ) -> str:
        """Usage docs:
        https://docs.pydantic.dev/2.10/concepts/serialization/#modelmodel_dump_json

        Generates a JSON representation of the model using Pydantic's `to_json` method.

        Args:
            indent: Indentation to use in the JSON output.
                If None is passed, the output will be compact.
            include: Field(s) to include in the JSON output.
            exclude: Field(s) to exclude from the JSON output.
            context: Additional context to pass to the serializer.
            by_alias: Whether to serialize using field aliases.
            exclude_unset: Whether to exclude fields that have not been explicitly set.
            exclude_defaults: Whether to exclude fields that are set to
                their default value.
            exclude_none: Whether to exclude fields that have a value of `None`.
            round_trip: If True, dumped values should be valid as input
                for non-idempotent types such as Json[T].
            warnings: How to handle serialization errors. False/"none" ignores them,
                True/"warn" logs errors, "error" raises a
                [`PydanticSerializationError`][pydantic_core.PydanticSerializationError].
            serialize_as_any: Whether to serialize fields with duck-typing serialization
                behavior.

        Returns:
            A JSON string representation of the model.
        """
        d = json.loads(
            BaseModel.model_dump_json(
                self,
                indent=indent,
                include=include,
                exclude=exclude,
                context=context,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=False,  # handle None values separately
                round_trip=round_trip,
                warnings=warnings,
                serialize_as_any=serialize_as_any,
            )
        )  # ToDo directly use dict?
        # this may replace some None values with IRIs in case they were never resolved
        # thats why we handle exclude_none there
        self._object_to_iri(d)
        if exclude_none:
            d = self.remove_none(d)
        return json.dumps(d, **dumps_kwargs)

    def to_jsonld(self) -> Dict:
        """Return the RDF representation of the object as JSON-LD."""
        return export_jsonld(self, BaseModel)

    @classmethod
    def from_jsonld(self, jsonld: Dict) -> "LinkedBaseModel":
        """Constructs a model instance from a JSON-LD representation."""
        return import_jsonld(BaseModel, jsonld, _types)



from typing import Callable, Dict

import pyld
from pydantic import BaseModel
from pydantic.v1 import BaseModel as BaseModel_v1
from pyld import jsonld


class GenericLinkedBaseModel:
    def _object_to_iri(self, d, exclude_none=False):
        for name in list(d.keys()):  # force copy of keys for inline-delete
            if name in self.__iris__:
                d[name] = self.__iris__[name]
            if exclude_none and d[name] is None:
                del d[name]
        return d

    @staticmethod
    def remove_none(d: Dict) -> Dict:
        """Remove None values from a dictionary recursively."""
        if isinstance(d, dict):
            return {
                k: GenericLinkedBaseModel.remove_none(v)
                for k, v in d.items()
                if v is not None
            }
        elif isinstance(d, list):
            return [GenericLinkedBaseModel.remove_none(i) for i in d]
        else:
            return d


def get_jsonld_context_loader(model_cls, model_type) -> Callable:
    """to overwrite the default jsonld document loader to load
    relative context from the osl"""

    classes = [model_cls]
    i = 0
    while 1:
        try:
            cls = classes[i]
            if cls == model_type:
                break
        except IndexError:
            break
        i += 1
        classes[i:i] = [base for base in cls.__bases__ if base not in classes]

    schemas = {}
    for base_class in classes:
        schema = {}
        if model_type == BaseModel:
            if hasattr(base_class, "model_config"):
                schema = base_class.model_config.get("json_schema_extra", {})
        if model_type == BaseModel_v1:
            if hasattr(base_class, "__config__"):
                schema = base_class.__config__.schema_extra
        id = schema.get("iri", None)
        schemas[id] = schema

    # print(schemas)

    def loader(url, options=None):
        if options is None:
            options = {}
        # print("Requesting", url)
        if url in schemas:
            schema = schemas[url]

            doc = {
                "contentType": "application/json",
                "contextUrl": None,
                "documentUrl": url,
                "document": schema,
            }
            # print("Loaded", doc)
            return doc

        else:
            requests_loader = pyld.documentloader.requests.requests_document_loader()
            return requests_loader(url, options)

    return loader


def export_jsonld(model_instance, model_type) -> Dict:
    """Return the RDF representation of the object as JSON-LD."""
    if model_type == BaseModel:
        # get the context from self.ConfigDict.json_schema_extra["@context"]
        context = model_instance.model_config.get("json_schema_extra", {}).get(
            "@context", {}
        )
    if model_type == BaseModel_v1:
        context = model_instance.__class__.__config__.schema_extra.get("@context", {})
    data = model_instance.dict()
    if "id" not in data and "@id" not in data:
        data["id"] = model_instance.get_iri()
    jsonld_dict = {"@context": context, **data}
    jsonld.set_document_loader(
        get_jsonld_context_loader(model_instance.__class__, model_type)
    )
    jsonld_dict = jsonld.expand(jsonld_dict)
    if isinstance(jsonld_dict, list):
        jsonld_dict = jsonld_dict[0]
    return jsonld_dict


def import_jsonld(model_type, jsonld_dict: Dict, _types: Dict[str, type]):
    """Return the object instance from the JSON-LD representation."""
    # ToDo: apply jsonld frame with @id restriction
    # get the @type from the jsonld_dict
    type_iri = jsonld_dict.get("@type", None)
    # if type_iri is None, return None
    if type_iri is None:
        return None
    # if type_iri is a list, get the first element
    if isinstance(type_iri, list):
        type_iri = type_iri[0]
    # get the class from the _types dict
    # Todo: IRI normalization
    type_iri = type_iri.split("/")[-1]
    model_cls = _types.get(type_iri, None)
    # if model_type is None, return None
    if model_cls is None:
        return None
    if model_type == BaseModel:
        # get the context from self.ConfigDict.json_schema_extra["@context"]
        context = model_cls.model_config.get("json_schema_extra", {}).get(
            "@context", {}
        )
    if model_type == BaseModel_v1:
        context = model_cls.__config__.schema_extra.get("@context", {})
    jsonld.set_document_loader(get_jsonld_context_loader(model_cls, model_type))
    jsonld_dict = jsonld.compact(jsonld_dict, context)
    if "@context" in jsonld_dict:
        del jsonld_dict["@context"]
    return model_cls(**jsonld_dict)



import json
import os
import re
from pathlib import Path
from tempfile import TemporaryDirectory

import datamodel_code_generator
from datamodel_code_generator import DataModelType, InputFileType, generate

from oold.utils.codegen import OOLDJsonSchemaParser


class Generator:
    def _generate(
        self,
        json_schemas,
        main_schema=None,
        output_model_type=DataModelType.PydanticV2BaseModel,
    ):
        # monkey patch class
        datamodel_code_generator.parser.jsonschema.JsonSchemaParser = (
            OOLDJsonSchemaParser
        )

        with TemporaryDirectory() as temporary_directory_name:
            temporary_directory = Path(temporary_directory_name)
            temporary_directory = Path(__file__).parent / "model" / "src"

            for schema in json_schemas:
                name = schema["id"]
                os.makedirs(
                    os.path.dirname(Path(temporary_directory / (name + ".json"))),
                    exist_ok=True,
                )
                with open(
                    Path(temporary_directory / (name + ".json")), "w", encoding="utf-8"
                ) as f:
                    schema_str = json.dumps(
                        schema, ensure_ascii=False, indent=4
                    ).replace("dollarref", "$ref")
                    # print(schema_str)
                    f.write(schema_str)

            input = Path(temporary_directory)
            output = Path(__file__).parent / "generated"
            if main_schema is not None:
                input = Path(temporary_directory / Path(main_schema))
                output = Path(__file__).parent / "model" / "model.py"

            if output_model_type == DataModelType.PydanticV2BaseModel:
                base_class = "oold.model.LinkedBaseModel"
            else:
                base_class = "oold.model.v1.LinkedBaseModel"
            generate(
                input_=input,
                # json_schema,
                input_file_type=InputFileType.JsonSchema,
                # input_filename="Foo.json",
                output=output,
                # set up the output model types
                output_model_type=output_model_type,
                # custom_template_dir=Path(model_dir_path),
                field_include_all_keys=True,
                base_class=base_class,
                # use_default = True,
                enum_field_as_literal="all",
                use_title_as_name=True,
                use_schema_description=True,
                use_field_description=True,
                encoding="utf-8",
                use_double_quotes=True,
                collapse_root_models=True,
                reuse_model=True,
                # create MyEnum(str, Enum) instead of MyEnum(Enum)
                use_subclass_enum=True,
                additional_imports=["pydantic.ConfigDict"]
                if output_model_type == DataModelType.PydanticV2BaseModel
                else [],
                apply_default_values_for_required_fields=True,
            )

            if main_schema is not None:
                content = ""
                with open(output, "r", encoding="utf-8") as f:
                    content = f.read()
                os.remove(output)

                content = re.sub(
                    r"(UUID = Field\(...)",
                    r"UUID = Field(default_factory=uuid4",
                    content,
                )  # enable default value for uuid

                if output_model_type == DataModelType.PydanticBaseModel:
                    # we are now using pydantic.v1
                    # pydantic imports lead to uninitialized fields
                    # (FieldInfo still present)
                    content = re.sub(
                        r"(from pydantic import)", "from pydantic.v1 import", content
                    )

                # write the content to the file
                with open(output, "w", encoding="utf-8") as f:
                    f.write(content)

    def preprocess(self, json_schemas):
        for schema in json_schemas:
            # schema = self.merge_property_schemas(schema)
            for property_key in schema.get("properties", {}):
                property = schema["properties"][property_key]
                if "range" in property:
                    if "type" in property:
                        del property["type"]
                    if "format" in property:
                        del property["format"]
                    # if range is a string we create a allOf with a ref to the range
                    if isinstance(property["range"], str):
                        property["allOf"] = [{"$ref": property["range"]}]
                    else:
                        property["$ref"] = property["range"]
                    if "required" in schema and property_key in schema["required"]:
                        # if no default value is set, remove the property from required
                        if "default" not in property:
                            schema["required"].remove(property_key)
                        if "x-oold-required-iri" not in property:
                            property["x-oold-required-iri"] = True
                if "items" in property:
                    if "range" in property["items"]:
                        if "type" in property["items"]:
                            del property["items"]["type"]
                        if "format" in property["items"]:
                            del property["items"]["format"]
                        if isinstance(property["items"]["range"], str):
                            property["items"]["allOf"] = [
                                {"$ref": property["items"]["range"]}
                            ]
                        else:
                            property["items"]["$ref"] = property["items"]["range"]
                        property["range"] = property["items"]["range"]
                        if "required" in schema and property_key in schema["required"]:
                            # if no default value is set,
                            # remove the property from required
                            if "default" not in property["items"]:
                                schema["required"].remove(property_key)
                            if "x-oold-required-iri" not in property:
                                property["x-oold-required-iri"] = True

                    if "properties" in property["items"]:
                        self.preprocess([property["items"]])

                if "properties" in property:
                    self.preprocess([property])

    def generate(
        self,
        json_schemas,
        main_schema=None,
        output_model_type=DataModelType.PydanticV2BaseModel,
    ):
        # pprint(json_schemas)
        self.preprocess(json_schemas)
        # pprint(json_schemas)
        self._generate(json_schemas, main_schema, output_model_type)
